import abc
import inspect
from typing import Callable, Generic, ParamSpec, TypeVar, get_args, get_origin

from haven.type_inspector import is_dataclass_type
from haven.types import Dataclass
from haven.utils import ParsingError

TConf = TypeVar("TConf", bound=Dataclass, covariant=True)
TRet = TypeVar("TRet")

P = ParamSpec("P")


class Component(Generic[TConf, TRet]):
    """A simple wrapper that couples a config instance with a function or Class which
    accepts that config as the first argument.
    """

    def __init__(self, config: TConf, func: Callable[..., TRet]):
        self.config = config
        self.func = func

    def __call__(self, *args, **kwargs) -> TRet:
        """Call the underlying function with the config dataclass as the first argument.
        All other args are forwarded."""
        return self.func(self.config, *args, **kwargs)


class ComponentBuilder(abc.ABC):
    @abc.abstractmethod
    def get_dataclass(self) -> type[Dataclass]: ...

    @abc.abstractmethod
    def build(self, conf: Dataclass) -> Component: ...


def get_component_dataclass(obj) -> type[Dataclass]:
    if is_dataclass_type(obj):
        return obj
    if isinstance(obj, ComponentBuilder):
        return obj.get_dataclass()
    arg_offset = 0
    if inspect.isclass(obj):
        obj = obj.__init__
        arg_offset = 1
    if inspect.isfunction(obj):
        params = list(inspect.signature(obj).parameters.values())
        if len(params) <= arg_offset:
            raise ParsingError(
                f"Choice function '{obj.__name__}' does not take any arguments, expeceted 1"
            )

        argdc = params[arg_offset].annotation
        if not is_dataclass_type(argdc):
            raise ParsingError(
                f"Choice function '{obj.__name__}' first argument's type is not a dataclass"
            )
        return argdc
    else:
        raise ValueError()


def get_component_type_dataclass(obj) -> type[Dataclass]:
    type_args = get_args(obj)
    if len(type_args) != 2:
        raise ParsingError("Component type should have two type args")
    return type_args[0]


def build_component(
    config: TConf, value: Callable[..., TRet] | ComponentBuilder
) -> Component[TConf, TRet]:
    if isinstance(value, ComponentBuilder):
        return value.build(config)
    return Component(config, value)


def is_component_type(ty) -> bool:
    orig = get_origin(ty)
    if orig is None:
        return False
    return issubclass(orig, Component)



"""Utilities for inspecting type hints, class types, and run-time types generically.

Based on:
  https://github.com/ilevkivskyi/typing_inspect/tree/master (MIT License)
  https://github.com/lebrice/SimpleParsing/blob/master/simple_parsing/utils.py (MIT License)

Cleaned up for python 3.10+ only.
"""

import builtins
import collections.abc
import dataclasses
import inspect
import types
import typing
from enum import Enum
from typing import (
    Any,
    Container,
    Dict,
    ForwardRef,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    Set,
    Tuple,
    TypeGuard,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from haven.types import Dataclass

T = TypeVar("T")

builtin_types = [
    getattr(builtins, d) for d in dir(builtins) if isinstance(getattr(builtins, d), type)
]


def is_typevar(t) -> bool:
    return type(t) is TypeVar


def get_bound(t):
    if is_typevar(t):
        return getattr(t, "__bound__", None)
    else:
        raise TypeError(f"type is not a `TypeVar`: {t}")


def is_forward_ref(t) -> TypeGuard[typing.ForwardRef]:
    return isinstance(t, typing.ForwardRef)


def get_forward_arg(fr: ForwardRef) -> str:
    return getattr(fr, "__forward_arg__")


def is_dataclass_instance(obj: Any) -> TypeGuard[Dataclass]:
    return dataclasses.is_dataclass(obj) and dataclasses.is_dataclass(type(obj))


def is_dataclass_type(obj: Any) -> TypeGuard[type[Dataclass]]:
    return inspect.isclass(obj) and dataclasses.is_dataclass(obj)


def get_container_item_type(container_type: type[Container[T]]) -> Any:
    """Returns the `type` of the items in the provided container `type`.

    When no type annotation is found, or no item type is found, returns
    `typing.Any`.
    NOTE: If a type with multiple arguments is passed, only the first type
    argument is returned.

    >>> import typing
    >>> from typing import List, Tuple
    >>> get_container_item_type(list)
    typing.Any
    >>> get_container_item_type(List)
    typing.Any
    >>> get_container_item_type(tuple)
    typing.Any
    >>> get_container_item_type(Tuple)
    typing.Any
    >>> get_container_item_type(List[int])
    <class 'int'>
    >>> get_container_item_type(List[str])
    <class 'str'>
    >>> get_container_item_type(List[float])
    <class 'float'>
    >>> get_container_item_type(List[float])
    <class 'float'>
    >>> get_container_item_type(List[Tuple])
    typing.Tuple
    >>> get_container_item_type(List[Tuple[int, int]])
    typing.Tuple[int, int]
    >>> get_container_item_type(Tuple[int, str])
    <class 'int'>
    >>> get_container_item_type(Tuple[str, int])
    <class 'str'>
    >>> get_container_item_type(Tuple[str, str, str, str])
    <class 'str'>

    Arguments:
        list_type {Type} -- A type, preferably one from the Typing module (List, Tuple, etc).

    Returns:
        Type -- the type of the container's items, if found, else Any.
    """
    if container_type in {
        list,
        set,
        tuple,
        List,
        Set,
        Tuple,
        Dict,
        Mapping,
        MutableMapping,
    }:
        # the built-in `list` and `tuple` types don't have annotations for their item types.
        return Any
    type_arguments = getattr(container_type, "__args__", None)
    if type_arguments:
        return type_arguments[0]
    else:
        return Any


def _mro(t: type) -> Sequence[type]:
    # TODO: This is mostly used in 'is_tuple' and such, and should be replaced with
    # either the built-in 'get_origin' from typing, or from typing-inspect.
    if t is None:
        return []
    if hasattr(t, "__mro__"):
        return t.__mro__
    elif get_origin(t) is type:
        return []
    elif hasattr(t, "mro") and callable(t.mro):
        return t.mro()
    return []


def is_list(t: type) -> bool:
    """returns True when `t` is a List type.

    Args:
        t (Type): a type.

    Returns:
        bool: True if `t` is list or a subclass of list.

    >>> from typing import *
    >>> is_list(list)
    True
    >>> is_list(tuple)
    False
    >>> is_list(List)
    True
    >>> is_list(List[int])
    True
    >>> is_list(List[Tuple[int, str, None]])
    True
    >>> is_list(Optional[List[int]])
    False
    >>> class foo(List[int]):
    ...     pass
    >>> is_list(foo)
    True
    """
    return list in _mro(t)


def is_tuple(t: type) -> bool:
    """returns True when `t` is a tuple type.

    Args:
        t (Type): a type.

    Returns:
        bool: True if `t` is tuple or a subclass of tuple.

    >>> from typing import *
    >>> is_tuple(list)
    False
    >>> is_tuple(tuple)
    True
    >>> is_tuple(Tuple)
    True
    >>> is_tuple(Tuple[int])
    True
    >>> is_tuple(Tuple[int, str, None])
    True
    >>> class foo(tuple):
    ...     pass
    >>> is_tuple(foo)
    True
    >>> is_tuple(List[int])
    False
    """
    return tuple in _mro(t)


def is_dict(t: type) -> bool:
    """returns True when `t` is a dict type or annotation.

    Args:
        t (Type): a type.

    Returns:
        bool: True if `t` is dict or a subclass of dict.

    >>> from typing import *
    >>> from collections import OrderedDict
    >>> is_dict(dict)
    True
    >>> is_dict(OrderedDict)
    True
    >>> is_dict(tuple)
    False
    >>> is_dict(Dict)
    True
    >>> is_dict(Dict[int, float])
    True
    >>> is_dict(Dict[Any, Dict])
    True
    >>> is_dict(Optional[Dict])
    False
    >>> is_dict(Mapping[str, int])
    True
    >>> class foo(Dict):
    ...     pass
    >>> is_dict(foo)
    True
    """
    mro = _mro(t)
    return dict in mro or Mapping in mro or collections.abc.Mapping in mro


def is_set(t: type) -> bool:
    """returns True when `t` is a set type or annotation.

    Args:
        t (Type): a type.

    Returns:
        bool: True if `t` is set or a subclass of set.

    >>> from typing import *
    >>> is_set(set)
    True
    >>> is_set(Set)
    True
    >>> is_set(tuple)
    False
    >>> is_set(Dict)
    False
    >>> is_set(Set[int])
    True
    >>> is_set(Set["something"])
    True
    >>> is_set(Optional[Set])
    False
    >>> class foo(Set):
    ...     pass
    >>> is_set(foo)
    True
    """
    return set in _mro(t)


def is_dataclass_type_or_typevar(t: type) -> bool:
    """Returns whether t is a dataclass type or a TypeVar of a dataclass type.

    Args:
        t (Type): Some type.

    Returns:
        bool: Whether its a dataclass type.
    """
    return dataclasses.is_dataclass(t) or (is_typevar(t) and dataclasses.is_dataclass(get_bound(t)))


def is_enum(t: type) -> bool:
    if inspect.isclass(t):
        return issubclass(t, Enum)
    return Enum in _mro(t)


def is_bool(t: type) -> bool:
    return bool in _mro(t)


def is_tuple_or_list(t: type) -> bool:
    return is_list(t) or is_tuple(t)


def is_union(t: type) -> bool:
    """Returns whether or not the given Type annotation is a variant (or subclass) of typing.Union.

    Args:
        t (Type): some type annotation

    Returns:
        bool: Whether this type represents a Union type.

    >>> from typing import *
    >>> is_union(Union[int, str])
    True
    >>> is_union(Union[int, str, float])
    True
    >>> is_union(Tuple[int, str])
    False
    """
    if isinstance(t, types.UnionType):
        return True
    return getattr(t, "__origin__", "") == Union


def is_homogeneous_tuple_type(t: type[tuple]) -> bool:
    """Returns whether the given Tuple type is homogeneous: if all items types are the same.

    This also includes Tuple[<some_type>, ...]

    Returns
    -------
    bool

    >>> from typing import *
    >>> is_homogeneous_tuple_type(Tuple)
    True
    >>> is_homogeneous_tuple_type(Tuple[int, int])
    True
    >>> is_homogeneous_tuple_type(Tuple[int, str])
    False
    >>> is_homogeneous_tuple_type(Tuple[int, str, float])
    False
    >>> is_homogeneous_tuple_type(Tuple[int, ...])
    True
    >>> is_homogeneous_tuple_type(Tuple[Tuple[int, str], ...])
    True
    >>> is_homogeneous_tuple_type(Tuple[List[int], List[str]])
    False
    """
    if not is_tuple(t):
        return False
    type_arguments = get_type_arguments(t)
    if not type_arguments:
        return True
    assert isinstance(type_arguments, tuple), type_arguments
    if len(type_arguments) == 2 and type_arguments[1] is Ellipsis:
        return True
    # Tuple[str, str, str] -> True
    # Tuple[str, str, float] -> False
    # TODO: Not sure if this will work with more complex item times (like nested tuples)
    return len(set(type_arguments)) == 1


def is_optional(t: type) -> bool:
    """Returns True if the given Type is a variant of the Optional type.

    Parameters
    ----------
    - t : Type

        a Type annotation (or "live" type)

    Returns
    -------
    bool
        Whether or not this is an Optional.

    >>> from typing import Union, Optional, List
    >>> is_optional(str)
    False
    >>> is_optional(Optional[str])
    True
    >>> is_optional(Union[str, None])
    True
    >>> is_optional(Union[str, List])
    False
    >>> is_optional(Union[str, List, int, float, None])
    True
    """
    return is_union(t) and type(None) in get_type_arguments(t)


def is_tuple_or_list_of_dataclasses(t: type) -> bool:
    return is_tuple_or_list(t) and is_dataclass_type_or_typevar(get_container_item_type(t))


def get_type_arguments(container_type: type) -> tuple[type, ...]:
    return get_args(container_type)


def get_type_name(some_type: type):
    result = getattr(some_type, "__name__", str(some_type))
    type_arguments = get_type_arguments(some_type)
    if type_arguments:
        result += f"[{','.join(get_type_name(T) for T in type_arguments)}]"
    return result



from .api import (
    add_decoder,
    add_encoder,
    dump,
    dump_pytree,
    load,
    load_pytree,
    update,
    update_from_dotlist,
)
from .choice import choice, plugin
from .component import Component, ComponentBuilder
from .utils import ParsingError

__all__ = [
    "add_decoder",
    "add_encoder",
    "choice",
    "dump",
    "dump_pytree",
    "load",
    "load_pytree",
    "ParsingError",
    "update_from_dotlist",
    "update",
    "plugin",
    "Component",
    "ComponentBuilder",
]

__version__ = "0.0.2"



"""Extensible dataclass to pytree encoding.

Use the external API function `add_encoder` to register encoders for new types.

import haven
import numpy as np
@haven.add_encoder
def encode_ndarray(obj: np.ndarray) -> str:
    return obj.tostring()
"""

from argparse import Namespace
from collections.abc import Hashable, Mapping
from dataclasses import fields, is_dataclass
from enum import Enum
from functools import singledispatch
from logging import getLogger
from os import PathLike
from typing import Any

from haven.component import Component
from haven.utils import ParsingError

logger = getLogger(__name__)


@singledispatch
def encode(obj: Any) -> Any:
    """Encodes an object into a json/yaml-compatible primitive type."""
    if is_dataclass(obj):
        d: dict[str, Any] = dict()
        for field in fields(obj):
            value = getattr(obj, field.name)
            try:
                d[field.name] = encode(value)
            except TypeError as e:
                raise ParsingError(f"Unable to encode field {field.name}") from e
        return d
    elif obj is None:
        return None
    else:
        raise ParsingError(
            f"No parser for object {obj} of type {type(obj)}, consider using haven.encode.register"
        )


@encode.register(Mapping)
def encode_dict(obj: Mapping) -> dict[Any, Any] | list[tuple[Any, Any]]:
    result = {}
    for k, v in obj.items():
        k_ = encode(k)
        v_ = encode(v)
        if isinstance(k_, Hashable):
            result[k_] = v_
        else:
            return [(encode(x), encode(y)) for x, y in obj.items()]
    return result


@encode.register(Enum)
def encode_enum(obj: Enum) -> str:
    return obj.name


@encode.register(Component)
def encode_component(obj: Component) -> dict[str, Any]:
    return encode(obj.config)


for t in [str, float, int, bool, bytes]:
    encode.register(t, lambda x: x)

for t in [list, tuple, set]:
    encode.register(t, lambda x: list(map(encode, x)))

encode.register(PathLike, lambda x: x.__fspath__())

encode.register(Namespace, lambda x: encode(vars(x)))



from dataclasses import dataclass

from mylib.config import Config, TaskConfig


@dataclass
class TrainConfig(TaskConfig):
    num_workers: int = 4


def run(train_cfg: TrainConfig, cfg: Config):
    print(f"Training with {train_cfg.num_workers} workers!")

    model = cfg.model()
    print(model)



from dataclasses import dataclass

import haven


@dataclass
class TaskConfig:
    name: str


class Model:
    pass


@dataclass
class ModelConfig:
    name: str


@dataclass
class Config:
    task: haven.Component[TaskConfig, None] = haven.plugin(
        discover_packages_path="mylib.tasks",
        attr="run",
    )
    model: haven.Component[ModelConfig, Model] = haven.choice(
        ["mylib.models.GPT", "mylib.models.Mamba"],
    )
    out_dir: str = "./output"



import argparse
import sys

import haven
from mylib.config import Config


def parse_args(argv):
    parser = argparse.ArgumentParser("my_program")
    parser.add_argument("--cfg", required=True, help="Path to config file to load.")
    parser.add_argument("overrides", nargs=argparse.REMAINDER)
    return parser.parse_args(argv)


def main(arg_list):
    args = parse_args(arg_list)

    # Load yaml config file and cli overrides
    with open(args.cfg, "r") as f:
        cfg = haven.load(Config, f, dotlist_overrides=args.overrides)

    # Print yaml
    print(haven.dump(cfg))

    cfg.task(cfg)


if __name__ == "__main__":
    main(sys.argv[1:])



from collections.abc import Callable
from typing import Any, Optional, TypeVar

from haven import formats
from haven.decoding import decode, type_decoders
from haven.encoding import encode
from haven.types import Dataclass, Pytree, ReadStream
from haven.utils import pytree_merge

TDataclass = TypeVar("TDataclass", bound=Dataclass)
T = TypeVar("T")


def load_pytree(cls: type[TDataclass], pytree: Pytree) -> TDataclass:
    """Decode a pytree into an instance of the given dataclass.

    Args:
        cls (type[TDataclass]): Dataclass to decode.
        pytree (Pytree): A tree of dicts and basic python objects matching the structure of `cls`.

    Returns:
        TDataclass: an instance of `cls` with values given by `pytree`.
    """
    return decode(cls, pytree)


def dump_pytree(obj: Dataclass) -> Pytree:
    """Encode a dataclass instance as a pytree

    Args:
        obj (Any): the datalass instance to convert.

    Returns:
        Pytree: a tree of basic python types matching the given dataclass.
    """
    return encode(obj)


def update(dc: TDataclass, overrides: Pytree) -> TDataclass:
    """Update a dataclass given overrides as a Pytree.

    Returns a copy and does not modify the original dataclass.

    Example:

      .. code-block:: python

        @dataclass
        Config:
            val: a = 5
            val2: b = 6

        cfg = haven.load_pytree(Config, {"a": 10})
        cfg = haven.update(cfg, {"b": 10})

        assert cfg.a == 10 and cf.b == 10

    Args:
        dc (TDataclass): Dataclass instance to update
        overrides (Pytree): Overrides as a pytree

    Returns:
        TDataclass: A new dataclass with overrides applied
    """
    pytree = encode(dc)
    pytree_merge(pytree, overrides)
    return decode(dc.__class__, pytree)


def update_from_dotlist(dc: TDataclass, dotlist: list[str]) -> TDataclass:
    """Update a dataclass instance with the overrides specified in the dotlist.

    Dotlists are lists of key=val strings. The keys can use dot syntax to reference
    fields deeper in the config tree. The values are parsed as yaml is. Example:

        .. code-block:: python
            update_from_dotlist(
                my_class, ["optim.lr=0.1", "model.arch=v5", "num_workers=6"]
            )

    This functionality is useful for providing command-line overrides. This function does
    not modify the original instance but instead returns a new copy.

    Args:
        dc (TDataclass): Dataclass instance to update.
        dotlist (list[str]): List of keyvalue strings.

    Returns
        TDataclass: the updated dataclass, as a copy.
    """
    pytree = formats.encode_dotlist(dotlist)
    return update(dc, pytree)


def add_encoder(ty: type[T], func: Callable[[T], Any]) -> Callable[[T], Any]:
    """Add function for converting instances of the given type to a pytree.

    Pytrees are composed of basic python builtin-types that are supported by
    YAML, JSON, and other similar formats.

    Args:
        ty (type[T]): the type to register
        func (Callable[[T], Any]): a function for encoding instances of the type T into a pytree.
    """
    encode.register(ty, func)
    return func


def add_decoder(
    ty: type[T], func: Callable[[Any], T], include_subclasses: bool = False
) -> Callable[[Any], T]:
    """Add a function for converting pytrees to instances of the given type.

    Pytrees are composed of basic python builtin-types that are supported by
    YAML, JSON, and other similar formats.

    Args:
        ty (type[T]): the type to register
        func (Callable[[Any], T]): a function that decodes pytrees into instances of T.
        include_subclasses (bool, optional): Whether to also apply this decoder for all subclasses of T. Defaults to False.
    """
    type_decoders.register(ty, func, include_subclasses)
    return func


def load(
    cls: type[TDataclass],
    stream: ReadStream,
    format: str = "yaml",
    overrides: Optional[Pytree] = None,
    dotlist_overrides: Optional[list[str]] = None,
) -> TDataclass:
    """Parse a text-based markup format and load it into an instance of the given dataclass

    Supported formats:
     - "yaml"
     - "json"

    Args:
        cls (type[TDataclass]): A dataclass to instantiate
        stream (ReadStream): A string or file-like object to read
        format (str, optional): The format to expect. Defaults to "yaml".
        overrides (Pytree, optional): A pytree of values to override the values from the stream.
        dotlist_overrides(list[str], optional): A list of overrides in "dotlist format", e.g.
            `["model.num_layers=5", "optim.lr=1e-3", "something.sizes=[1,2,3]"]`. Useful in
            combination with argparse.REMAINDER to provide CLI config setting.

    Raises:
        ValueError: If an unknown format is specified.

    Returns:
        TDataclass: An instance of the given dataclass with values taken from the text document.
    """
    if format == "yaml":
        pytree = formats.encode_yaml(stream)
    elif format == "json":
        pytree = formats.encode_json(stream)
    else:
        raise ValueError(f"Invalid format '{format}'. Expected one of yaml, json.")

    if overrides is not None:
        pytree_merge(pytree, overrides)

    if dotlist_overrides is not None:
        dotlist_pytree = formats.encode_dotlist(dotlist_overrides)
        pytree_merge(pytree, dotlist_pytree)

    return decode(cls, pytree)


def dump(obj: Dataclass, format: str = "yaml") -> str:
    """Convert an instance of a dataclass

    See :meth:`load` for a list of supported formats.

    Args:
        obj (Dataclass): the object to encode.
        format (str, optional): the format to output. Defaults to "yaml".

    Raises:
        ValueError: If an unknown format is specified.

    Returns:
        str: a text representation of the dataclass in the specified format.
    """
    pytree = encode(obj)
    if format == "yaml":
        return formats.decode_yaml(pytree)
    elif format == "json":
        return formats.decode_json(pytree)
    else:
        raise ValueError("Invalid format '{format}'. Expeceted one of yaml, json.")



from typing import Any

from haven.types import Pytree, ReadStream
from haven.utils import deflatten


def encode_yaml(stream: ReadStream) -> Any:
    """
    Read yaml from a string or file-like object and encode it into a pytree.
    """
    if isinstance(stream, (str, bytes, bytearray)) and not stream:
        return {}

    import yaml

    return yaml.safe_load(stream)


def encode_json(stream: ReadStream) -> Any:
    """
    Read json from a string or file-like object and encode it into a pytree.
    """
    import json

    if isinstance(stream, (str, bytes, bytearray, memoryview)):
        if not stream:
            return {}
        return json.loads(stream)

    return json.load(stream)


def decode_yaml(obj: Any) -> str:
    """
    Convert a pytree into yaml string.
    """
    import yaml

    return yaml.dump(obj)


def decode_json(obj: Any) -> str:
    """
    Convert a pytree into a json string.
    """
    import json

    return json.dumps(obj)


def encode_dotlist(dotlist: list[str]) -> Pytree:
    """
    Parse a dotlist into a pytree.
    """
    kv = {}
    for arg in dotlist:
        idx = arg.find("=")
        if idx == -1:
            raise ValueError(f"dotlist argument '{arg}' invalid: no key.")
        if idx == len(arg) - 1:
            raise ValueError(f"dotlist argument '{arg}' invalid: no value given")
        if idx == 0:
            raise ValueError(f"dotlist argument '{arg} invalid: empty key")
        key = arg[0:idx]
        value = encode_yaml(arg[idx + 1 :])
        kv[key] = value

    return deflatten(kv)



from typing import TYPE_CHECKING, Any, TypeAlias

if TYPE_CHECKING:
    from _typeshed import DataclassInstance as _DataclassInstance
    from _typeshed import SupportsRead


Dataclass: TypeAlias = "_DataclassInstance"

# File-like types supported by config parser libraries like pyyaml.
ReadStream: TypeAlias = "str | bytes | SupportsRead[str] | SupportsRead[bytes]"

Pytree: TypeAlias = dict[str, Any]



from dataclasses import dataclass

from mylib.config import Model, ModelConfig


@dataclass
class GPTConfig(ModelConfig):
    embed_dim: int = 512
    seq_len: int = 512
    head_dim: int = 64


class GPT(Model):
    def __init__(self, cfg: GPTConfig):
        self.seq_len = cfg.seq_len

    def __repr__(self):
        return f"<GPT seq_len={self.seq_len}>"


@dataclass
class MambaConfig(ModelConfig):
    hdim: int = 512
    nblock: int = 10


class Mamba(Model):
    def __init__(self, cfg: MambaConfig):
        self.num_blocks = cfg.nblock

    def __repr__(self):
        return f"<Mamba nblock={self.num_blocks}>"



"""Utility functions used in various parts of the haven package."""

import collections.abc as c_abc
import copy
import dataclasses
import logging
from collections.abc import Iterable, Mapping, MutableMapping
from logging import getLogger
from typing import (
    Any,
    TypeVar,
)

from haven.type_inspector import is_dataclass_type
from haven.types import Dataclass

logger = getLogger(__name__)

T = TypeVar("T")


class HavenException(Exception):
    pass


class ParsingError(HavenException):
    pass


def keep_keys(d: dict, keys_to_keep: Iterable[str]) -> tuple[dict, dict]:
    d_keys = set(d.keys())  # save a copy since we will modify the dict.
    removed = {}
    for key in d_keys:
        if key not in keys_to_keep:
            removed[key] = d.pop(key)
    return d, removed


def flatten(d, parent_key=None, sep="."):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, c_abc.MutableMapping):
            items.extend(flatten(v, parent_key=new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def deflatten(d: dict[str, Any], sep: str = "."):
    deflat_d = {}
    for key in d:
        key_parts = key.split(sep)
        curr_d = deflat_d
        for inner_key in key_parts[:-1]:
            if inner_key not in curr_d:
                curr_d[inner_key] = {}
            curr_d = curr_d[inner_key]
        curr_d[key_parts[-1]] = d[key]
    return deflat_d


def remove_matching(dict_a, dict_b):
    dict_a = flatten(dict_a)
    dict_b = flatten(dict_b)
    for key in dict_b:
        if key in dict_a and dict_b[key] == dict_a[key]:
            del dict_a[key]
    return deflatten(dict_a)


def format_error(e: Exception):
    try:
        return f"{type(e).__name__}: {e}"
    except Exception:
        return f"Exception: {e}"


def import_object(path: str):
    import importlib

    parts = path.rsplit(".", 1)
    if len(parts) != 2:
        raise ImportError(
            f"Failed to import object from '{path}'. Should have at least one '.' separating module from object."
        )
    mod, attr = parts
    mod = importlib.import_module(mod)
    if not hasattr(mod, attr):
        raise ImportError(f"Module '{mod}' has no attribute '{attr}")
    return getattr(mod, attr)


def pytree_merge(dest: MutableMapping[Any, Any], src: Mapping[Any, Any]) -> None:
    for src_k, src_v in src.items():
        if src_k in dest and isinstance(src_v, Mapping) and isinstance(dest[src_k], MutableMapping):
            pytree_merge(dest[src_k], src_v)
        else:
            dest[src_k] = copy.deepcopy(src_v)


def get_defaults_dict(c: type[Dataclass]):
    """ " Get defaults of a dataclass without generating the object"""
    defaults_dict = {}
    for field in dataclasses.fields(c):
        if is_dataclass_type(field.type):
            defaults_dict[field.name] = get_defaults_dict(field.type)
        elif field.default is not dataclasses.MISSING:
            defaults_dict[field.name] = field.default
        elif field.default_factory is not dataclasses.MISSING:
            try:
                defaults_dict[field.name] = field.default_factory()
            except Exception as e:
                logging.debug(
                    f"Failed getting default for field {field.name} using its default factory.\n\tUnderlying error: {e}"
                )
                continue
    return defaults_dict



"""Functions for decoding dataclass fields from "raw" values (e.g. from json)."""

from collections import OrderedDict
from collections.abc import Callable
from dataclasses import MISSING, Field, fields, is_dataclass
from functools import lru_cache, partial
from logging import getLogger
from pathlib import Path
from typing import (
    Any,
    Optional,
    TypeVar,
)

from haven.choice import ChoiceMeta
from haven.component import (
    build_component,
    get_component_dataclass,
    get_component_type_dataclass,
    is_component_type,
)
from haven.type_inspector import (
    get_bound,
    get_type_arguments,
    is_dataclass_type,
    is_dict,
    is_enum,
    is_list,
    is_set,
    is_tuple,
    is_typevar,
    is_union,
)
from haven.type_registry import RegistryFunc, TypeRegistry
from haven.types import Dataclass
from haven.utils import (
    ParsingError,
    format_error,
)

logger = getLogger(__name__)
T = TypeVar("T", bound=Dataclass)
K = TypeVar("K")
V = TypeVar("V")


type_decoders = TypeRegistry()


def decode(cls: type[T], raw_value: Any) -> T:
    return get_decoding_fn(cls)(raw_value)


@lru_cache(maxsize=128)
def get_decoding_fn(cls: type[T]) -> Callable[[Any], T]:
    """Fetches/Creates a decoding function for the given type annotation.

    This decoding function can then be used to create an instance of the type
    when deserializing dicts.
    """
    # Start by trying the dispatch mechanism
    registered_func: Optional[RegistryFunc] = type_decoders.dispatch(cls)
    if registered_func is not None:
        # If supports subclasses, pass the actual type
        if registered_func.include_subclasses:
            return partial(registered_func.func, cls)
        else:
            return registered_func.func

    if is_dataclass_type(cls):
        return partial(decode_dataclass, cls)  # type: ignore

    if cls is Any:
        logger.debug(f"Decoding an Any type: {cls}")
        return no_op

    if is_dict(cls):
        logger.debug(f"Decoding a Dict field: {cls}")
        args = get_type_arguments(cls)
        if args is None or len(args) != 2:
            args = (Any, Any)
        return decode_dict(*args)  # type: ignore

    if is_set(cls):
        logger.debug(f"Decoding a Set field: {cls}")
        args = get_type_arguments(cls)
        if args is None or len(args) != 1:
            args = (Any,)
        return decode_set(args[0])  # type: ignore

    if is_tuple(cls):
        logger.debug(f"Decoding a Tuple field: {cls}")
        args = get_type_arguments(cls)
        if args is None:
            args = []
        return decode_tuple(*args)  # type: ignore

    if is_list(cls):  # NOTE: Looks like can't be written with a dictionary
        logger.debug(f"Decoding a List field: {cls}")
        args = get_type_arguments(cls)
        if args is None or len(args) != 1:
            # Using a `List` or `list` annotation, so we don't know what do decode the
            # items into!
            args = (Any,)
        decode_fn = decode_list(args[0])  # type: ignore

        return decode_fn  # type: ignore

    if is_union(cls):
        logger.debug(f"Decoding a Union field: {cls}")
        args = get_type_arguments(cls)
        return decode_union(*args)

    if is_enum(cls):
        return lambda key: cls[key]  # type: ignore

    if is_typevar(cls):
        bound = get_bound(cls)
        logger.debug(f"Decoding a typevar: {cls}, bound type is {bound}.")
        if bound is not None:
            return get_decoding_fn(bound)

    raise ParsingError(f"No decoding function for type {cls}, consider using haven.decode.register")


def decode_dataclass(cls: type[Dataclass], d: dict[str, Any]) -> Dataclass:
    """Parses an instance of the dataclass `cls` from the dict `d`."""
    obj_dict: dict[str, Any] = d.copy()

    init_args: dict[str, Any] = {}
    non_init_args: dict[str, Any] = {}

    logger.debug(f"from_dict for {cls}")

    for field in fields(cls):
        name = field.name

        if "_haven_choices" in field.metadata:
            raw_value = obj_dict.pop(name, {})
            field_value = decode_choice_field(field, raw_value, cls, d)
        elif name not in obj_dict:
            if field.default is MISSING and field.default_factory is MISSING:
                logger.warning(
                    f"Couldn't find the field '{name}' in the dict with keys " f"{list(d.keys())}"
                )
            continue
        else:
            raw_value = obj_dict.pop(name)
            try:
                field_value = decode_field(field, raw_value)
            except ParsingError as e:
                raise e
            except Exception as e:
                raise ParsingError(
                    f'Failed when parsing value=\'{raw_value}\' into field "{cls}.{name}" of type {field.type}.\n\tUnderlying error is "{format_error(e)}"'
                )

        if field.init:
            init_args[name] = field_value
        else:
            non_init_args[name] = field_value

    extra_args = obj_dict

    # If there are arguments left over in the dict after taking all fields.
    if extra_args:
        raise Exception(f"The fields {extra_args} do not belong to the class")

    init_args.update(extra_args)
    try:
        instance = cls(**init_args)  # type: ignore
    except TypeError as e:
        raise ParsingError(
            f"Couldn't instantiate class {cls} using the given arguments.\n\t Underlying error: {e}"
        )

    for name, value in non_init_args.items():
        logger.debug(f"Setting non-init field '{name}' on the instance.")
        setattr(instance, name, value)
    return instance


def decode_field(field: Field, raw_value: Any) -> Any:
    """Converts a "raw" value (e.g. from json file) to the type of the `field`."""
    name = field.name
    field_type = field.type
    logger.debug(f"Decode name = {name}, type = {field_type}")
    return decode(field_type, raw_value)


def decode_choice_field(
    field: Field, raw_value: Any, outer_cls: type[Dataclass], outer_dict: dict[str, Any]
) -> Any:
    meta: ChoiceMeta = field.metadata["_haven_choices"]

    if not isinstance(raw_value, dict):
        raise ParsingError(
            f"Choice field '{field.name}' only works for dataclasses, expected dict when decoding"
        )

    targ_dict = outer_dict if meta.outer else raw_value
    choice_name = None

    if meta.key_field not in targ_dict:
        # Key field name not found in the dict, find the default value from the field definition
        targ_cls = outer_cls if meta.outer else field.type
        if is_component_type(targ_cls):
            targ_cls = get_component_type_dataclass(targ_cls)
        if not is_dataclass(targ_cls):
            raise ParsingError("The choice field type is not a dataclass.")
        for f in fields(targ_cls):
            if f.name == meta.key_field:
                if f.default is not MISSING:
                    choice_name = f.default
                elif f.default_factory is not MISSING:
                    choice_name = f.default_factory()

        if choice_name is None:
            raise ParsingError(
                f"Choice field '{field.name}' cannot find key field '{meta.key_field}' for determining which choice to use, and no default value was found."
            )
    else:
        choice_name = targ_dict[meta.key_field]

        if not isinstance(choice_name, str):
            raise ParsingError(f"Choice field '{field.name}' has '{meta.key_field}")

    choice = meta.choices.get_value(choice_name)

    if choice is None:
        raise ParsingError(
            f"Invalid choice '{choice_name}' for field '{field.name}'. Valid choices are: {','.join(meta.choices.get_choice_names())}"
        )

    cls = get_component_dataclass(choice)
    conf = decode_dataclass(cls, raw_value)

    if is_component_type(field.type):
        return build_component(conf, choice)
    elif is_dataclass_type(field.type):
        return conf
    else:
        raise ParsingError()


def decode_optional(t: type[T]) -> Callable[[Any | None], T | None]:
    decode = get_decoding_fn(t)

    def _decode_optional(val: Any | None) -> T | None:
        return val if val is None else decode(val)

    return _decode_optional


def try_functions(*funcs: Callable[[Any], T]) -> Callable[[Any], T | Any]:
    """Tries to use the functions in succession, else returns the same value unchanged."""

    def _try_functions(val: Any) -> T | Any:
        for func in funcs:
            try:
                return func(val)
            except Exception:
                continue
        # If no function worked, raise an exception
        raise ParsingError(f"No valid parsing for value {val}")

    return _try_functions


def decode_union(*types: type) -> Callable[[Any], Any]:
    types: list[type[Any]] = list(types)
    optional = type(None) in types
    # Partition the Union into None and non-None types.
    while type(None) in types:
        types.remove(type(None))

    decoding_fns: list[Callable[[Any], Any]] = [
        decode_optional(t) if optional else get_decoding_fn(t)  # type: ignore
        for t in types
    ]  # type: ignore
    # Try using each of the non-None types, in succession. Worst case, return the value.
    return try_functions(*decoding_fns)


def decode_list(t: type[T]) -> Callable[[list[Any]], list[T]]:
    decode_item = get_decoding_fn(t)

    def _decode_list(val: list[Any]) -> list[T]:
        # assert type(val) == list
        if not isinstance(val, list):
            raise ParsingError(f"The given value='{val}' is not of a valid input")
        return [decode_item(v) for v in val]

    return _decode_list


def decode_tuple(*tuple_item_types: type[T]) -> Callable[[list[T]], tuple[T, ...]]:
    """Makes a parsing function for creating tuples."""
    # Get the decoding function for each item type
    has_ellipsis = False
    if Ellipsis in tuple_item_types:
        # TODO: This isn't necessary, the ellipsis will always be at index 1.
        ellipsis_index = tuple_item_types.index(Ellipsis)
        decoding_fn_index = ellipsis_index - 1
        decoding_fn = get_decoding_fn(tuple_item_types[decoding_fn_index])
        has_ellipsis = True
    elif len(tuple_item_types) == 0:
        has_ellipsis = True
        decoding_fn = no_op  # Functionality will be the same as Tuple[Any,...]
    else:
        decoding_fns = [get_decoding_fn(t) for t in tuple_item_types]

    # Note, if there are more values than types in the tuple type, then the
    # last type is used.

    def _decode_tuple(val: list[T]) -> tuple[T, ...]:
        if val is None:
            raise TypeError("Value must not be None for conversion to a tuple")
        if has_ellipsis:
            return tuple(decoding_fn(v) for v in val)
        else:
            if len(decoding_fns) != len(val):
                err_msg = (
                    f"Trying to decode {len(val)} values for a predfined {len(decoding_fns)}-Tuple"
                )
                raise ParsingError(err_msg)
            return tuple(decoding_fns[i](v) for i, v in enumerate(val))

    return _decode_tuple


def decode_set(item_type: type[T]) -> Callable[[list[T]], set[T]]:
    """Makes a parsing function for creating sets with items of type `item_type`."""
    # Get the parsers fn for a list of items of type `item_type`.
    parse_list_fn = decode_list(item_type)

    def _decode_set(val: list[Any]) -> set[T]:
        return set(parse_list_fn(val))

    return _decode_set


def decode_dict(K_: type[K], V_: type[V]) -> Callable[[list[tuple[Any, Any]]], dict[K, V]]:
    """Creates a decoding function for a dict type. Works with OrderedDict too."""
    decode_k = get_decoding_fn(K_)
    decode_v = get_decoding_fn(V_)

    def _decode_dict(val: dict[Any, Any] | list[tuple[Any, Any]]) -> dict[K, V]:
        result: dict[K, V] = {}
        if isinstance(val, list):
            result = OrderedDict()
            items = val
        elif isinstance(val, OrderedDict):
            # NOTE: Needed to propagate `OrderedDict` type
            result = OrderedDict()
            items = val.items()
        else:
            items = val.items()
        for k, v in items:
            k_ = decode_k(k)
            v_ = decode_v(v)
            result[k_] = v_
        return result

    return _decode_dict


def no_op(v: T) -> T:
    """Decoding function that gives back the value as-is."""
    return v


def try_constructor(t: type[T]) -> Callable[[Any], T | Any]:
    """Tries to use the type as a constructor. If that fails, returns the value as-is."""
    return try_functions(lambda val: t(**val), lambda val: t(val))  # type: ignore


# Dictionary mapping from types/type annotations to their decoding functions.
for t in [str, float, int, bool, bytes]:
    type_decoders.register(t, t)

type_decoders.register(Path, Path)



import abc
import importlib
import pkgutil
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from functools import cache
from inspect import isclass, isfunction
from typing import Any, Optional, TypeAlias, TypeVar

from haven.type_inspector import is_dataclass_type
from haven.types import Dataclass
from haven.utils import ParsingError, format_error, import_object

T = TypeVar("T")
TDataclass = TypeVar("TDataclass", bound=Dataclass)
ChoiceType: TypeAlias = str | Callable[..., Any] | type[Dataclass]


def try_import(spec: str):
    try:
        cls: type[Dataclass] = import_object(spec)
    except Exception as e:
        raise ParsingError(
            f"Could not import choice at '{spec}'.\nUnderlying error: {format_error(e)}"
        )
    return cls


class ChoiceProvider(abc.ABC):
    @abc.abstractmethod
    def get_choices(self) -> dict[str, str | type[Dataclass]]: ...

    def get_value(self, choice_name: str) -> Optional[type[Dataclass]]:
        choices = self.get_choices()
        if choice_name not in choices:
            return None
        choice = choices[choice_name]
        if isinstance(choice, str):
            return try_import(choice)
        else:
            return choice

    def get_choice_names(self) -> Iterable[str]:
        return self.get_choices().keys()


class DictProvider(ChoiceProvider):
    def __init__(self, choices: dict[str, ChoiceType]):
        self.choices = choices

    def get_choices(self):
        return self.choices


def parse_choices(
    choices: list[ChoiceType] | dict[str, ChoiceType] | ChoiceProvider,
) -> ChoiceProvider:
    if isinstance(choices, (list, tuple)):
        return iterable_provider(choices)
    elif isinstance(choices, dict):
        return DictProvider(choices)
    elif isinstance(choices, ChoiceProvider):
        return choices
    else:
        raise TypeError("Choices must be list, dict, or ChoiceProvider.")


class Plugin(ChoiceProvider):
    def __init__(self, discover_packages_path: str, attr: str):
        self.discover_packages_path = discover_packages_path
        self.attr = attr

    @cache
    def get_choices(self) -> dict[str, str]:
        package_module = importlib.import_module(self.discover_packages_path, __package__)

        choices = {}
        for mod_info in pkgutil.iter_modules(package_module.__path__):
            module = importlib.import_module(f"{self.discover_packages_path}.{mod_info.name}")
            if hasattr(module, self.attr):
                val = getattr(module, self.attr)
                print(val)
                if isinstance(val, (list, dict, ChoiceProvider)):
                    for k, v in parse_choices(val).get_choices().items():
                        print(k, v)
                        if k in choices:
                            raise ParsingError(
                                f"Duplicate choices registered for '{k}' in plugin provider."
                            )
                        choices[k] = v
                else:
                    choices[mod_info.name] = val

        return choices


def iterable_provider(choices: Iterable[ChoiceType]) -> ChoiceProvider:
    choice_dict = {}
    for c in choices:
        if isinstance(c, str):
            parts = c.rsplit(".", 1)
            if len(parts) != 2:
                raise ValueError(
                    f"Choice strings must be valid object import path (e.g. module.MyClass) Got: '{c}'"
                )
            choice_dict[parts[-1]] = c
        elif isfunction(c) or is_dataclass_type(c) or isclass(c):
            choice_dict[c.__name__] = c
        else:
            raise TypeError(f"Objects passed as choices must be dataclasses. Got: '{c}'")

    return DictProvider(choice_dict)


@dataclass
class ChoiceMeta:
    choices: ChoiceProvider
    key_field: str = "name"
    outer: bool = False


def choice(
    choices: list[ChoiceType] | dict[str, ChoiceType] | ChoiceProvider,
    key_field: str = "name",
    outer: bool = False,
    default_factory: Callable[[], T] | str | None = None,
) -> T:
    """A field which dynamically selects the dataclass to decode based on the value of another string field.

    Choice fields construct a name, dataclass mapping. During config parsing, the value of another
    field in the child or parent dataclass is used to lookup which dataclass to decode.

    The `choices` argument can be either a list, dict, or instance of ChoiceProvider; if it is a list,
    then the name for each choice is determined automatically by inspecting the name of the class or function.
    The list/dict value can either contain the choices directly or strings of the form `my_package.my_module.MyClass`.

    Note:
        The `default_factory` argument works as expected in standard dataclass construction, but when constructed
        via this library the value of the key field will always override it, since it controls which choice to use.

    Args:
        choices (list[ChoiceType] | dict[str, ChoiceType] | ChoiceProvider): determines the possible choices
        key_field (str, optional): the name of the field which determines the choice at runtime. Defaults to "name".
        outer (bool, optional): If true, the choice field is a sibling in the parent dataclass. If false, it is a
            member of this field's dataclass. Defaults to False.
        default_factory (Callable[..., T] | str, optional): If str, import the object specified and use that class
            as the default factory. Otherwise, a callable that returns a dataclass instance or None for no default.

    Raises:
        ValueError: raised if import strings don't follow the correct format.
        TypeError: raised if the passed objects are not dataclasses or if `choices` is not a dict or list.
    """
    provider = parse_choices(choices)

    metadata = {
        "_haven_choices": ChoiceMeta(
            choices=provider,
            key_field=key_field,
            outer=outer,
        ),
    }

    kwargs = {}
    if isinstance(default_factory, str):
        kwargs["default_factory"] = import_object(default_factory)
    elif default_factory is not None:
        kwargs["default_factory"] = default_factory
    else:
        kwargs["default_factory"] = lambda: None

    return field(metadata=metadata, **kwargs)


def plugin(
    discover_packages_path: str,
    attr: str,
    key_field: str = "name",
    outer: bool = False,
    default_factory: Callable[[], T] | str | None = None,
):
    """Discover choices as all modules in a package namespace automatically.

    Similar to the :meth:`choices` function, but it doesn't require explicitly
    specifying all available choices upfront. Instead, all submodules of the given
    package are discovered at decoding time as available choices.

    Args:
        discover_packages_path (str): the namespace in which to discover packages
        attr (str): the attribute expected in each module to use as the choice value
        key_field (str, optional): The field use to determine the choice. Defaults to "name".
            Same function as in :meth:`choice`
        outer (bool, optional): Same as in :meth:`choice`. Defaults to False.
        default_factory (Callable[[], T] | str | None, optional): Defaults to None.
    """
    return choice(
        Plugin(discover_packages_path, attr),
        key_field,
        outer=outer,
        default_factory=default_factory,
    )
